import traceback
from functools import partial
from pathlib import Path

import click
import yaml
from tqdm import tqdm

from ccmm_invenio.conversion.base import VocabularyReader
from ccmm_invenio.conversion.ccmm_copy import CopyReader  # noqa
from ccmm_invenio.conversion.ccmm_csv import CSVReader  # noqa
from ccmm_invenio.conversion.ccmm_filtered import (  # noqa
    DescendantsOfFilter,
    FilteredReader,
)
from ccmm_invenio.conversion.ccmm_sparql import SPARQLReader  # noqa


@click.command()
@click.argument("vocabulary_names", nargs=-1)
def convert_vocabularies(vocabulary_names: list[str]) -> None:
    root_dir = Path(__file__).parent.parent

    converters: list[tuple[VocabularyReader, Path]] = [
        (
            # TODO: the CSV does not include dataCiteCode, we needd to add it there !!!
            CSVReader("Agent Role", root_dir / "input/CCMM_slovniky(AgentRole).csv"),
            root_dir / "fixtures/ccmm_agent_roles.yaml",
        ),
        (
            CSVReader(
                "Alternate Title",
                root_dir / "input/CCMM_slovniky(AlternateTitle).csv",
            ),
            root_dir / "fixtures/ccmm_alternate_title_types.yaml",
        ),
        (
            CSVReader(
                "Location Relation",
                root_dir / "input/CCMM_slovniky(LocationRelation).csv",
            ),
            root_dir / "fixtures/ccmm_location_relation_types.yaml",
        ),
        (
            CSVReader(
                "Relation Type",
                root_dir / "input/CCMM_slovniky(RelationType).csv",
                extra=[root_dir / "input/addon_relation_types.csv"],
            ),
            root_dir / "fixtures/ccmm_relation_types.yaml",
        ),
        (
            CSVReader(
                "Subject Category",
                root_dir / "input/CCMM_slovniky(SubjectCategory).csv",
            ),
            root_dir / "fixtures/ccmm_subject_categories.yaml",
        ),
        (
            CSVReader(
                "Time Reference",
                root_dir / "input/CCMM_slovniky(TimeReference).csv",
            ),
            root_dir / "fixtures/ccmm_time_reference_types.yaml",
        ),
        (
            SPARQLReader(
                "Languages",
                "http://publications.europa.eu/resource/authority/language",
                "http://publications.europa.eu/resource/authority/language",
                extra_props={
                    "ISO_639_2T": """
                        ?concept skos:notation ?ISO_639_2T FILTER(datatype(?ISO_639_2T) = euvoc:ISO_639_2T)
                    """,
                    "ISO_639_1": """
                        ?concept skos:notation ?ISO_639_1 FILTER(datatype(?ISO_639_1) = euvoc:ISO_639_1)
                    """,
                    "ISO_639_3": """
                        ?concept skos:notation ?ISO_639_3 FILTER(datatype(?ISO_639_3) = euvoc:ISO_639_3)
                    """,
                    "XML_LNG": """
                        ?concept skos:notation ?XML_LNG FILTER(datatype(?XML_LNG) = euvoc:XML_LNG)
                    """,
                    "ISO_639_2B": """
                        ?concept skos:notation ?ISO_639_2B FILTER(datatype(?ISO_639_2B) = euvoc:ISO_639_2B)
                    """,
                },
                prefixes={
                    "euvoc": "http://publications.europa.eu/ontology/euvoc#",
                },
            ),
            root_dir / "fixtures/ccmm_languages.yaml",
        ),
        (
            SPARQLReader(
                "Access Rights",
                "https://vocabularies.coar-repositories.org/access_rights/access_rights.nt",
                "http://purl.org/coar/access_right/scheme",
                format="turtle",
                load_subgraphs=False,
                extra=root_dir / "input/addon_access_rights.ttl",
            ),
            root_dir / "fixtures/ccmm_access_rights.yaml",
        ),
        (
            SPARQLReader(
                "Resource Type",
                "https://vocabularies.coar-repositories.org/resource_types/resource_types.nt",
                "http://purl.org/coar/resource_type/scheme",
                format="turtle",
                load_subgraphs=False,
                extra=root_dir / "input/addon_resource_types.ttl",
                extra_props={
                    "zenodo": """
                        ?concept props:zenodo ?zenodo
                    """,
                },
                prefixes={
                    "props": "http://vocabs.ccmm.cz/props/",
                },
                array_resolution=zenodo_resource_type_array_resolution,
            ),
            root_dir / "fixtures/ccmm_resource_types.yaml",
        ),
        (
            SPARQLReader(
                "File types",
                "http://publications.europa.eu/resource/authority/file-type",
                "http://publications.europa.eu/resource/authority/file-type",
            ),
            root_dir / "fixtures/ccmm_file_types.yaml",
        ),
        (
            CopyReader(
                "Licenses",
                root_dir / "input/licenses.yaml",
            ),
            root_dir / "fixtures/ccmm_licenses.yaml",
        ),
        (
            CopyReader(
                "Subject schemes",
                root_dir / "input/subject_schemes.yaml",
            ),
            root_dir / "fixtures/ccmm_subject_schemes.yaml",
        ),
        (
            FilteredReader(
                "Contributor type",
                root_dir / "fixtures/ccmm_agent_role.yaml",
                filter_cls=partial(
                    DescendantsOfFilter,
                    descendants_of={
                        "https://vocabs.ccmm.cz/registry/codelist/AgentRole/Contributor"
                    },
                ),
            ),
            root_dir / "fixtures/ccmm_contributor_types.yaml",
        ),
    ]

    if not vocabulary_names:
        vocabulary_names = [reader.name for reader, _ in converters]

    with_progress = tqdm(converters, leave=False, unit="vocab")
    for reader, output_path in with_progress:
        if reader.name not in vocabulary_names:
            continue
        try:
            with_progress.set_description(reader.name)
            with_progress.refresh()
            data = reader.data()
            with open(output_path, "w", encoding="utf-8") as output_file:
                yaml.safe_dump_all(
                    data,
                    output_file,
                    allow_unicode=True,
                    default_flow_style=False,
                )
        except Exception as e:
            print(f"Error converting {reader.name}: {e}")
            traceback.print_exc()


def zenodo_resource_type_array_resolution(prop: str, parent: dict[str, str]) -> None:
    values = parent[prop]
    parent[prop] = ", ".join(sorted(values))

    if prop == "zenodo":
        for value in values:
            parent[f"zenodo-{value}"] = "true"


if __name__ == "__main__":
    convert_vocabularies()



from pathlib import Path
from typing import Any, Callable

import yaml

from .base import VocabularyReader


class FilterCls:
    """Base class for filter classes."""

    def __init__(self):
        pass

    def filter(self, items: list[dict[str, Any]]) -> list[dict[str, Any]]:
        """Condition to filter items."""
        # Implement the filtering logic here
        return items


class FilteredReader(VocabularyReader):
    """Read and filter vocabulary data."""

    def __init__(
        self,
        name: str,
        input_file: Path,
        filter_cls: type[FilterCls] | Callable[[], FilterCls],
    ):
        """Initialize the reader."""
        super().__init__(name)
        self.input_file = input_file
        self.filter_cls = filter_cls

    def data(self):
        """Read the vocabulary data and apply the filter function."""
        with open(self.input_file, "r", encoding="utf-8") as file:
            _data = list(yaml.safe_load_all(file))

        # Apply the filter function
        flt = self.filter_cls()
        data = flt.filter(_data)

        # Remove hierarchy.parent from items that are not in the filtered list
        ids = {item["id"] for item in data}
        for item in data:
            if "hierarchy" in item and "parent" in item["hierarchy"]:
                parent = item["hierarchy"]["parent"]
                if parent not in ids:
                    del item["hierarchy"]

        return data


class DescendantsOfFilter(FilterCls):
    """Filter class to filter items based on descendants."""

    def __init__(self, descendants_of: set[str]):
        """Initialize the filter with a set of descendants."""
        super().__init__()
        self.descendants_of = descendants_of

    def filter(self, items: list[dict[str, Any]]) -> list[dict[str, Any]]:
        known_parents: set[str] = set()

        # register initial known parents
        for item in items:
            iri = item.get("props", {}).get("iri")
            if iri and iri in self.descendants_of:
                known_parents.add(item["id"])

        ret: list[dict[str, Any]] = []
        for item in items:
            parent = item.get("hierarchy", {}).get("parent")
            if not parent:
                continue
            if parent in known_parents:
                known_parents.add(item["id"])
                ret.append(item)

        return ret



from typing import Any


class VocabularyReader:
    """Base class for all readers."""

    def __init__(self, name: str):
        """Initialize the reader with a name."""
        self.name = name

    def data(self) -> list[dict[str, Any]]:
        """Read the data from the source."""
        raise NotImplementedError("Subclasses must implement this method.")



import csv
from pathlib import Path
from typing import Any

from .base import VocabularyReader


class CSVReader(VocabularyReader):
    """Read CCMM CSV files and convert them to YAML format."""

    def __init__(
        self, name: str, csv_path: Path, extra: list[Path] | None = None
    ) -> None:
        super().__init__(name)
        self.csv_path = csv_path
        self.extra = extra or []

    def data(self) -> list[dict[str, str]]:
        """Convert CCMM CSV to YAML that can be imported to NRP Invenio."""

        with open(self.csv_path, "r", encoding="utf-8-sig") as csv_file:
            reader = csv.DictReader(csv_file, delimiter=";", quotechar='"')
            rows = list(reader)

        # Remove leading and trailing whitespace from all keys and values
        converted_data: list[dict[str, str]] = []
        for row in rows:
            row = {key.strip(): value.strip() for key, value in row.items() if key}

            # IRI;base IRI;parentId;id;title_cs;title_en;definition_cs;definition_en
            term_id = row.pop("id")
            iri = row.pop("IRI")
            base_iri = row.pop("base IRI")
            parent_id = row.pop("parentId")
            title_cs = row.pop("title_cs")
            title_en = row.pop("title_en")
            definition_cs = row.pop("definition_cs")
            definition_en = row.pop("definition_en")

            if not term_id or (not title_cs and not title_en):
                # Skip empty rows
                continue

            term: dict[str, Any] = {
                "id": term_id,
                "title": {
                    "cs": title_cs,
                    "en": title_en,
                },
                "description": {
                    "cs": definition_cs,
                    "en": definition_en,
                },
                "props": {
                    "iri": iri,
                    "base_iri": base_iri,
                },
            }
            if parent_id:
                term["hierarchy"] = {
                    "parent": parent_id,
                }
            converted_data.append(term)

        converted_data_by_id = {term["id"]: term for term in converted_data}
        for extra_file in self.extra:
            with open(extra_file, "r", encoding="utf-8-sig") as extra_csv_file:
                extra_reader = csv.DictReader(
                    extra_csv_file, delimiter=";", quotechar='"'
                )
                for extra_row in extra_reader:
                    extra_row = {
                        key.strip(): value.strip()
                        for key, value in extra_row.items()
                        if key
                    }
                    term_id = extra_row.pop("id")
                    if term_id in converted_data_by_id:
                        converted = converted_data_by_id[term_id]
                        for k, v in extra_row.items():
                            # k might contain a dot that means nesting
                            _set(converted, k, v)
        return converted_data


def _set(d: dict[str, Any], key: str, value: str) -> None:
    """Set a value in a nested dictionary based on a dot-separated key."""
    keys = key.split(".")
    for k in keys[:-1]:
        d = d.setdefault(k, {})
    d[keys[-1]] = value



import logging
from collections import defaultdict
from functools import partial
from pathlib import Path
from typing import Any, Callable, cast
from urllib.error import HTTPError, URLError

from rdflib import SKOS, Graph, URIRef
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)
from tqdm.contrib.concurrent import process_map

from .base import VocabularyReader

# Set up logging to see retry attempts
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Custom retry configuration
RETRY_CONFIG: dict[str, Any] = {
    "stop": stop_after_attempt(5),
    "wait": wait_exponential(multiplier=1, min=2, max=30),  # 2s, 4s, 8s, etc. up to 30s
    "retry": retry_if_exception_type((URLError, HTTPError, ConnectionError)),
    "before_sleep": before_sleep_log(logger, logging.WARNING),
    "reraise": True,
}


@retry(**RETRY_CONFIG)
def parse_source(url: str, format: str = "xml") -> Graph:
    """Parse RDF with automatic retries for transient errors."""
    g = Graph()
    try:
        g.parse(url, format=format)
        return g
    except Exception as e:
        logger.warning(f"Attempt failed for {url}: {str(e)}")
        raise  # Re-raise for tenacity to handle


def join_with_commas(prop: str, parent: dict[str, Any]) -> None:
    parent[prop] = ", ".join(sorted(parent[prop]))


class SPARQLReader(VocabularyReader):
    """Download rdf locally, enrich it and call sparql to convert the RDF to Invenio YAML format."""

    def __init__(
        self,
        name: str,
        endpoint: str,
        skos_concept: str,
        extra: Path | None = None,  # turtle serialization of extra triples
        load_subgraphs: bool = True,
        format: str = "xml",
        extra_props: dict[str, str] | None = None,
        prefixes: dict[str, str] | None = None,
        array_resolution: Callable[[str, dict[str, Any]], None] = join_with_commas,
    ):
        """Initialize the SPARQL reader.

        :param name: Name of the vocabulary.
        :param endpoint: SPARQL endpoint URL.
        :param skos_concept: SKOS concept scheme URI.
        :param extra: Path to a turtle file with extra triples.
        :param load_subgraphs: Whether to load subgraphs.
        :param format: Format of the SPARQL file at the endpoint.
        :param extra_props: Additional properties to include in the results.
                            Keys are the names of the properties,
                            and values is the sparql query to get the value,
                            will be wrapped in an OPTIONAL clause.
        :param prefixes: Prefixes to use in the SPARQL query.
        :param array_resolution: Function to resolve array properties (default is to join with commas).
        """
        super().__init__(name)
        self.endpoint = endpoint
        self.skos_concept = skos_concept
        self.extra = extra
        self.load_subgraphs = load_subgraphs
        self.format = format
        self.extra_props = extra_props or {}
        self.prefixes = prefixes or {}
        self.array_resolution = array_resolution

    def data(self) -> list[dict[str, str]]:
        """Convert CCMM from SPARQL to YAML that can be imported to NRP Invenio."""

        # whole_graph: Graph = parse_source("file:///tmp/ccmm.ttl", format="turtle")
        whole_graph: Graph = parse_source(self.endpoint, format=self.format)

        if self.load_subgraphs:
            self._load_subgraphs(whole_graph)

        # add extra triples to the graph
        if self.extra:
            extra_graph = Graph()
            with open(self.extra, "r", encoding="utf-8") as extra_file:
                extra_graph.parse(extra_file, format="turtle")
            whole_graph += extra_graph

        # # save the whole graph to a turtle file /tmp/ccmm.ttl
        # whole_graph.serialize("/tmp/ccmm.ttl", format="turtle")

        query = """
    PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
    PREFIX dc: <http://purl.org/dc/elements/1.1/>
    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    {{prefixes}}

    SELECT ?concept ?label_cs ?label_en ?description_cs ?description_en ?identifier ?broader {{extra_props}}
    WHERE {
    ?concept a skos:Concept ;
                skos:inScheme ?scheme .

    # Get English label (prefLabel first, then altLabel)
    OPTIONAL {
        { ?concept skos:prefLabel ?prefLabel_en FILTER(lang(?prefLabel_en) = "en") }
        UNION
        { ?concept skos:altLabel ?altLabel_en }
        BIND(COALESCE(?prefLabel_en, ?altLabel_en) AS ?label_en)
    }
    
    # Get Czech label (prefLabel first, then english label)
    OPTIONAL {
        { ?concept skos:prefLabel ?label_cs FILTER(lang(?label_cs) = "cs") }
    }

    # Get English description (skos:definition)
    OPTIONAL {
        { ?concept skos:definition ?description_en FILTER(lang(?description_en) = "en") }
    }
    
    # Get Czech description (skos:definition)
    OPTIONAL {
        { ?concept skos:definition ?description_cs FILTER(lang(?description_cs) = "cs") }
    }

    # Get dc:identifier if available
    OPTIONAL { ?concept dc:identifier ?identifier }

    # Get skos:broader if available
    OPTIONAL { ?concept skos:broader ?broader }

    # Get extra properties
    {{extra_props_sparql}}

    }
    
    ORDER BY ?concept                                            
                        """

        query = query.replace(
            "{{prefixes}}",
            "\n".join(
                f"PREFIX {key}: <{value}>" for key, value in self.prefixes.items()
            ),
        )
        query = query.replace(
            "{{extra_props}}",
            " ".join(f"?{key}" for key in self.extra_props),
        )
        query = query.replace(
            "{{extra_props_sparql}}",
            "\n".join(f"OPTIONAL {{ {value} }}" for value in self.extra_props.values()),
        )
        rows = whole_graph.query(
            query,
            initBindings={
                "scheme": URIRef(self.skos_concept),
            },
        )
        converted: dict[str, Any] = {}
        by_iri: dict[str, str] = {}
        for row in cast(tuple[Any, ...], rows):
            row = [str(x) if x is not None else None for x in row]
            (
                iri,
                title_cs,
                title_en,
                description_cs,
                description_en,
                term_id,
                broader,
                *row_props,
            ) = row
            # Skip empty rows
            if not iri:
                continue
            # Skip rows without titles
            if title_cs is None and title_en is None:
                continue

            # set term_id if it was not provided
            if not term_id:
                term_id = iri.strip("/").split("/")[-1]
                term_id = term_id.strip("#").split("#")[-1]

            if term_id not in converted:
                term: dict[str, Any] = {"id": term_id}
                term["props"] = defaultdict(set[str])
                term["title"] = {}
                term["description"] = {}
                converted[term_id] = term
            else:
                term = converted[term_id]

            if "cs" not in term["title"]:
                term["title"]["cs"] = title_cs or title_en
            if "en" not in term["title"]:
                term["title"]["en"] = title_en or title_cs
            if "iri" not in term["props"]:
                term["props"]["iri"] = {iri}
            if "cs" not in term["description"] and description_cs:
                term["description"]["cs"] = description_cs
            if "en" not in term["description"] and description_en:
                term["description"]["en"] = description_en

            for prop, value in zip(self.extra_props.keys(), row_props):
                if value is not None:
                    term["props"][prop].add(str(value))

            if broader:
                term["hierarchy"] = {"parent": broader}

            by_iri[iri] = term_id

        # Resolve the hierarchy
        for term in converted.values():
            if "hierarchy" in term:
                parent = term["hierarchy"]["parent"]
                term["hierarchy"]["parent"] = by_iri[parent]

        # Convert sets to comma-separated strings
        for term in converted.values():
            term["props"] = dict(term["props"])
            for prop in list(term["props"]):
                self.array_resolution(prop, term["props"])

        to_sort = [dict(term) for term in converted.values()]

        ret_ids: set[str] = set()
        ret: list[dict[str, Any]] = []

        remaining: list[dict[str, Any]] = []
        for term in to_sort:
            # if there is no parent, add the term
            if "hierarchy" not in term:
                ret.append(term)
                ret_ids.add(term["id"])
                continue
            else:
                remaining.append(term)

        while remaining:
            to_sort = remaining
            remaining = []

            for term in to_sort:
                parent = term["hierarchy"]["parent"]
                if parent in ret_ids:
                    ret.append(term)
                    ret_ids.add(term["id"])
                else:
                    remaining.append(term)
            if len(to_sort) == len(remaining):
                raise ValueError(
                    f"There is a cycle in the hierarchy, please check the data: {remaining}"
                )

        return ret

    def _load_subgraphs(self, whole_graph: Graph) -> None:
        """Load all subgraphs from the SKOS concept scheme and merge them into the whole graph."""

        # Get all concepts
        scheme = URIRef(self.skos_concept)
        subjects = [str(x) for x in whole_graph.subjects(SKOS.inScheme, scheme)]

        # enrich graph with all the subjects
        for subject_graph in process_map(
            partial(parse_source, format=self.format),
            subjects,
            max_workers=20,
            chunksize=10,
            leave=False,
            unit="subgraph",
        ):
            whole_graph += cast(Graph, subject_graph)



from pathlib import Path

import yaml

from .base import VocabularyReader


class CopyReader(VocabularyReader):
    """Copy reader for copying vocabularies."""

    def __init__(self, name: str, input_file: Path):
        """Initialize the CopyReader."""
        super().__init__(name)
        self.input_file = input_file

    def data(self) -> list[dict[str, str]]:
        """Read the vocabulary from the input file."""
        with open(self.input_file, "r") as f:
            return list(yaml.safe_load_all(f))
